@@ -19,11 +19,11 @@ import os
 from unittest import mock
 import yaml
 
+import fixtures
 from oslo_config import cfg
 from oslo_context import context
 from oslo_serialization import jsonutils
 from oslotest import base as test_base
-import six
 
 from oslo_policy import _cache_handler
 from oslo_policy import _checks
@@ -238,13 +238,6 @@ class EnforcerTest(base.PolicyBaseTestCase):
         super(EnforcerTest, self).setUp()
         self.create_config_file('policy.json', POLICY_JSON_CONTENTS)
 
-    def check_loaded_files(self, filenames):
-        self.assertEqual(
-            [self.get_config_file_fullname(n)
-             for n in filenames],
-            self.enforcer._loaded_files
-        )
-
     def _test_scenario_with_opts_registered(self, scenario, *args, **kwargs):
         # This test registers some rules, calls the scenario and then checks
         # the registered rules. The scenario should be a method which loads
@@ -291,11 +284,6 @@ class EnforcerTest(base.PolicyBaseTestCase):
         loaded_rules = jsonutils.loads(str(self.enforcer.rules))
         self.assertEqual('role:fakeB', loaded_rules['default'])
         self.assertEqual('is_admin:True', loaded_rules['admin'])
-        self.check_loaded_files([
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-            os.path.join('policy.d', 'b.conf'),
-        ])
 
     def test_load_directory_after_file_update(self):
         self.create_config_file(
@@ -305,10 +293,6 @@ class EnforcerTest(base.PolicyBaseTestCase):
         loaded_rules = jsonutils.loads(str(self.enforcer.rules))
         self.assertEqual('role:fakeA', loaded_rules['default'])
         self.assertEqual('is_admin:True', loaded_rules['admin'])
-        self.check_loaded_files([
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-        ])
         new_policy_json_contents = jsonutils.dumps({
             "default": "rule:admin",
             "admin": "is_admin:True",
@@ -332,12 +316,41 @@ class EnforcerTest(base.PolicyBaseTestCase):
         self.assertEqual('role:fakeA', loaded_rules['default'])
         self.assertEqual('is_admin:True', loaded_rules['admin'])
         self.assertEqual('rule:bar', loaded_rules['foo'])
-        self.check_loaded_files([
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-        ])
+
+    def test_load_directory_after_file_is_emptied(self):
+        def dict_rules(enforcer_rules):
+            """Converts enforcer rules to dictionary.
+
+            :param enforcer_rules: enforcer rules represented as a class Rules
+            :return: enforcer rules represented as a dictionary
+            """
+            return jsonutils.loads(str(enforcer_rules))
+
+        self.assertEqual(self.enforcer.rules, {})
+
+        self.enforcer.load_rules()
+        main_policy_file_rules = jsonutils.loads(POLICY_JSON_CONTENTS)
+        self.assertEqual(main_policy_file_rules,
+                         dict_rules(self.enforcer.rules))
+
+        folder_policy_file = os.path.join('policy.d', 'a.conf')
+        self.create_config_file(folder_policy_file, POLICY_A_CONTENTS)
+        self.enforcer.load_rules()
+        expected_rules = main_policy_file_rules.copy()
+        expected_rules.update(jsonutils.loads(POLICY_A_CONTENTS))
+        self.assertEqual(expected_rules, dict_rules(self.enforcer.rules))
+
+        self.create_config_file(folder_policy_file, '{}')
+        # Force the mtime change since the unit test may write to this file
+        # too fast for mtime to actually change.
+        absolute_folder_policy_file_path = self.get_config_file_fullname(
+            folder_policy_file)
+        stinfo = os.stat(absolute_folder_policy_file_path)
+        os.utime(absolute_folder_policy_file_path,
+                 (stinfo.st_atime + 42, stinfo.st_mtime + 42))
+        self.enforcer.load_rules()
+        self.assertEqual(main_policy_file_rules,
+                         dict_rules(self.enforcer.rules))
 
     def test_load_directory_opts_registered(self):
         self._test_scenario_with_opts_registered(self.test_load_directory)
@@ -349,8 +362,7 @@ class EnforcerTest(base.PolicyBaseTestCase):
         self.enforcer.load_rules(False)
         self.assertIsNotNone(self.enforcer.rules)
 
-        old = six.next(six.itervalues(
-            self.enforcer._policy_dir_mtimes))
+        old = next(iter(self.enforcer._policy_dir_mtimes))
         self.assertEqual(1, len(self.enforcer._policy_dir_mtimes))
 
         # Touch the file
@@ -361,16 +373,10 @@ class EnforcerTest(base.PolicyBaseTestCase):
 
         self.enforcer.load_rules(False)
         self.assertEqual(1, len(self.enforcer._policy_dir_mtimes))
-        self.assertEqual(old, six.next(six.itervalues(
-            self.enforcer._policy_dir_mtimes)))
+        self.assertEqual(old, next(iter(self.enforcer._policy_dir_mtimes)))
 
         loaded_rules = jsonutils.loads(str(self.enforcer.rules))
         self.assertEqual('is_admin:True', loaded_rules['admin'])
-        self.check_loaded_files([
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-            os.path.join('policy.d', 'a.conf'),
-        ])
 
     def test_load_directory_caching_with_files_updated_opts_registered(self):
         self._test_scenario_with_opts_registered(
@@ -385,21 +391,15 @@ class EnforcerTest(base.PolicyBaseTestCase):
         self.enforcer.load_rules(False)
         self.assertIsNotNone(self.enforcer.rules)
 
-        old = six.next(six.itervalues(
-            self.enforcer._policy_dir_mtimes))
+        old = next(iter(self.enforcer._policy_dir_mtimes))
         self.assertEqual(1, len(self.enforcer._policy_dir_mtimes))
 
         self.enforcer.load_rules(False)
         self.assertEqual(1, len(self.enforcer._policy_dir_mtimes))
-        self.assertEqual(old, six.next(six.itervalues(
-            self.enforcer._policy_dir_mtimes)))
+        self.assertEqual(old, next(iter(self.enforcer._policy_dir_mtimes)))
 
         loaded_rules = jsonutils.loads(str(self.enforcer.rules))
         self.assertEqual('is_admin:True', loaded_rules['admin'])
-        self.check_loaded_files([
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-        ])
 
     def test_load_directory_caching_with_files_same_but_overwrite_false(self):
         self.test_load_directory_caching_with_files_same(overwrite=False)
@@ -457,12 +457,6 @@ class EnforcerTest(base.PolicyBaseTestCase):
         loaded_rules = jsonutils.loads(str(self.enforcer.rules))
         self.assertEqual('role:fakeC', loaded_rules['default'])
         self.assertEqual('is_admin:True', loaded_rules['admin'])
-        self.check_loaded_files([
-            'policy.json',
-            os.path.join('policy.d', 'a.conf'),
-            os.path.join('policy.d', 'b.conf'),
-            os.path.join('policy.2.d', 'fake.conf'),
-        ])
 
     def test_load_multiple_directories_opts_registered(self):
         self._test_scenario_with_opts_registered(
@@ -478,8 +472,6 @@ class EnforcerTest(base.PolicyBaseTestCase):
         self.assertIsNotNone(self.enforcer.rules)
         self.assertIn('default', self.enforcer.rules)
         self.assertIn('admin', self.enforcer.rules)
-        self.check_loaded_files(
-            ['policy.json', os.path.join('policy.d', 'a.conf')])
 
     def test_load_non_existed_directory_opts_registered(self):
         self._test_scenario_with_opts_registered(
@@ -788,6 +780,22 @@ class EnforcerTest(base.PolicyBaseTestCase):
                           policy.RuleDefault(name='owner',
                                              check_str='role:owner'))
 
+    def test_enforcer_does_not_modify_original_registered_rule(self):
+        rule_original = policy.RuleDefault(
+            name='test',
+            check_str='role:owner',)
+
+        self.enforcer.register_default(rule_original)
+        self.enforcer.registered_rules['test']._check_str = 'role:admin'
+        self.enforcer.registered_rules['test']._check = 'role:admin'
+
+        self.assertEqual(
+            self.enforcer.registered_rules['test'].check_str, 'role:admin')
+        self.assertEqual(
+            self.enforcer.registered_rules['test'].check, 'role:admin')
+        self.assertEqual(rule_original.check_str, 'role:owner')
+        self.assertEqual(rule_original.check.__str__(), 'role:owner')
+
     def test_non_reversible_check(self):
         self.create_config_file('policy.json',
                                 jsonutils.dumps(
@@ -869,23 +877,6 @@ class EnforcerTest(base.PolicyBaseTestCase):
         for k, v in expected_creds.items():
             self.assertEqual(expected_creds[k], creds[k])
 
-    @mock.patch('warnings.warn', new=mock.Mock())
-    def test_map_context_attributes_populated_system(self):
-        request_context = context.RequestContext(system_scope='all')
-        expected_creds = request_context.to_policy_values()
-        expected_creds['system'] = 'all'
-
-        creds = self.enforcer._map_context_attributes_into_creds(
-            request_context
-        )
-
-        # We don't use self.assertDictEqual here because to_policy_values
-        # actaully returns a non-dict object that just behaves like a
-        # dictionary, but does some special handling when people access
-        # deprecated policy values.
-        for k, v in expected_creds.items():
-            self.assertEqual(expected_creds[k], creds[k])
-
     def test_enforcer_accepts_policy_values_from_context(self):
         rule = policy.RuleDefault(name='fake_rule', check_str='role:test')
         self.enforcer.register_default(rule)
@@ -906,6 +897,20 @@ class EnforcerTest(base.PolicyBaseTestCase):
         target_dict = {}
         self.enforcer.enforce('fake_rule', target_dict, ctx)
 
+    def test_enforcer_understands_system_scope_creds_dict(self):
+        self.conf.set_override('enforce_scope', True, group='oslo_policy')
+        rule = policy.RuleDefault(
+            name='fake_rule', check_str='role:test', scope_types=['system']
+        )
+        self.enforcer.register_default(rule)
+
+        ctx = context.RequestContext()
+        creds = ctx.to_dict()
+        creds['system_scope'] = 'all'
+
+        target_dict = {}
+        self.enforcer.enforce('fake_rule', target_dict, creds)
+
     def test_enforcer_raises_invalid_scope_with_system_scope_type(self):
         self.conf.set_override('enforce_scope', True, group='oslo_policy')
         rule = policy.RuleDefault(
@@ -918,15 +923,23 @@ class EnforcerTest(base.PolicyBaseTestCase):
         target_dict = {}
         self.assertRaises(
             policy.InvalidScope, self.enforcer.enforce, 'fake_rule',
-            target_dict, ctx
+            target_dict, ctx, do_raise=True
         )
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(
+                'fake_rule', target_dict, ctx, do_raise=False))
 
         # model a project-scoped token, which should fail enforcement
         ctx = context.RequestContext(project_id='fake')
         self.assertRaises(
             policy.InvalidScope, self.enforcer.enforce, 'fake_rule',
-            target_dict, ctx
+            target_dict, ctx, True
         )
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(
+                'fake_rule', target_dict, ctx, do_raise=False))
 
     def test_enforcer_understands_domain_scope(self):
         self.conf.set_override('enforce_scope', True, group='oslo_policy')
@@ -951,15 +964,23 @@ class EnforcerTest(base.PolicyBaseTestCase):
         target_dict = {}
         self.assertRaises(
             policy.InvalidScope, self.enforcer.enforce, 'fake_rule',
-            target_dict, ctx
+            target_dict, ctx, True
         )
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(
+                'fake_rule', target_dict, ctx, do_raise=False))
 
         # model a project-scoped token, which should fail enforcement
         ctx = context.RequestContext(project_id='fake')
         self.assertRaises(
             policy.InvalidScope, self.enforcer.enforce, 'fake_rule',
-            target_dict, ctx
+            target_dict, ctx, True
         )
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(
+                'fake_rule', target_dict, ctx, do_raise=False))
 
     def test_enforcer_understands_project_scope(self):
         self.conf.set_override('enforce_scope', True, group='oslo_policy')
@@ -984,28 +1005,48 @@ class EnforcerTest(base.PolicyBaseTestCase):
         target_dict = {}
         self.assertRaises(
             policy.InvalidScope, self.enforcer.enforce, 'fake_rule',
-            target_dict, ctx
+            target_dict, ctx, True
         )
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(
+                'fake_rule', target_dict, ctx, do_raise=False))
 
         # model a domain-scoped token, which should fail enforcement
         ctx = context.RequestContext(domain_id='fake')
         self.assertRaises(
             policy.InvalidScope, self.enforcer.enforce, 'fake_rule',
-            target_dict, ctx
+            target_dict, ctx, True
         )
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(
+                'fake_rule', target_dict, ctx, do_raise=False))
+
+    def test_enforce_scope_with_subclassed_checks_when_scope_not_set(self):
+        self.conf.set_override('enforce_scope', True, group='oslo_policy')
+        rule = _checks.TrueCheck()
+        rule.scope_types = None
+        ctx = context.RequestContext(system_scope='all', roles=['admin'])
+        self.enforcer.enforce(rule, {}, ctx)
+
+    def test_enforcer_raises_invalid_scope_with_subclassed_checks(self):
+        self.conf.set_override('enforce_scope', True, group='oslo_policy')
+        rule = _checks.TrueCheck()
+        rule.scope_types = ['domain']
+        ctx = context.RequestContext(system_scope='all', roles=['admin'])
+        self.assertRaises(
+            policy.InvalidScope,
+            self.enforcer.enforce, rule, {}, ctx, do_raise=True)
+        # and the same should return False if do_raise=False
+        self.assertFalse(
+            self.enforcer.enforce(rule, {}, ctx, do_raise=False))
 
 
 class EnforcerNoPolicyFileTest(base.PolicyBaseTestCase):
     def setUp(self):
         super(EnforcerNoPolicyFileTest, self).setUp()
 
-    def check_loaded_files(self, filenames):
-        self.assertEqual(
-            [self.get_config_file_fullname(n)
-             for n in filenames],
-            self.enforcer._loaded_files
-        )
-
     def test_load_rules(self):
         # Check that loading rules with no policy file does not error
         self.enforcer.load_rules(True)
@@ -1031,10 +1072,6 @@ class EnforcerNoPolicyFileTest(base.PolicyBaseTestCase):
         loaded_rules = jsonutils.loads(str(self.enforcer.rules))
         self.assertEqual('role:fakeB', loaded_rules['default'])
         self.assertEqual('is_admin:True', loaded_rules['admin'])
-        self.check_loaded_files([
-            'policy.d/a.conf',
-            'policy.d/b.conf',
-        ])
 
 
 class CheckFunctionTestCase(base.PolicyBaseTestCase):
@@ -1238,7 +1275,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_deprecate_a_policy_check_string(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:create_bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N'
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1247,8 +1286,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         enforcer = policy.Enforcer(self.conf)
         enforcer.register_defaults(rule_list)
@@ -1278,7 +1315,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_deprecate_an_empty_policy_check_string(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:create_bar',
-            check_str=''
+            check_str='',
+            deprecated_reason='because of reasons',
+            deprecated_since='N',
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1287,8 +1326,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='because of reasons',
-            deprecated_since='N'
         )]
         enforcer = policy.Enforcer(self.conf)
         enforcer.register_defaults(rule_list)
@@ -1306,7 +1343,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_deprecate_replace_with_empty_policy_check_string(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:create_bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='because of reasons',
+            deprecated_since='N',
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1315,8 +1354,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='because of reasons',
-            deprecated_since='N'
         )]
         enforcer = policy.Enforcer(self.conf)
         enforcer.register_defaults(rule_list)
@@ -1333,15 +1370,7 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_deprecate_a_policy_name(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:bar',
-            check_str='role:baz'
-        )
-
-        rule_list = [policy.DocumentedRuleDefault(
-            name='foo:create_bar',
             check_str='role:baz',
-            description='Create a bar.',
-            operations=[{'path': '/v1/bars/', 'method': 'POST'}],
-            deprecated_rule=deprecated_rule,
             deprecated_reason=(
                 '"foo:bar" is not granular enough. If your deployment has '
                 'overridden "foo:bar", ensure you override the new policies '
@@ -1350,7 +1379,15 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
                 '"foo:bar:update", "foo:bar:list", and "foo:bar:delete", '
                 'which might be backwards incompatible for your deployment'
             ),
-            deprecated_since='N'
+            deprecated_since='N',
+        )
+
+        rule_list = [policy.DocumentedRuleDefault(
+            name='foo:create_bar',
+            check_str='role:baz',
+            description='Create a bar.',
+            operations=[{'path': '/v1/bars/', 'method': 'POST'}],
+            deprecated_rule=deprecated_rule,
         )]
         expected_msg = (
             'Policy "foo:bar":"role:baz" was deprecated in N in favor of '
@@ -1424,7 +1461,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_deprecate_check_str_suppress_does_not_log_warning(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:create_bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N'
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1433,8 +1472,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         enforcer = policy.Enforcer(self.conf)
         enforcer.suppress_deprecation_warnings = True
@@ -1446,7 +1483,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_deprecate_name_suppress_does_not_log_warning(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:bar',
-            check_str='role:baz'
+            check_str='role:baz',
+            deprecated_reason='"foo:bar" is not granular enough.',
+            deprecated_since='N',
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1455,8 +1494,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars/', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"foo:bar" is not granular enough.',
-            deprecated_since='N'
         )]
 
         rules = jsonutils.dumps({'foo:bar': 'role:bang'})
@@ -1494,7 +1531,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
     def test_suppress_default_change_warnings_flag_not_log_warning(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:create_bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N',
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1503,8 +1542,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         enforcer = policy.Enforcer(self.conf)
         enforcer.suppress_default_change_warnings = True
@@ -1513,7 +1550,7 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             enforcer.load_rules()
             mock_warn.assert_not_called()
 
-    def test_deprecated_policy_for_removal_must_include_deprecated_since(self):
+    def test_deprecated_policy_for_removal_must_include_deprecated_meta(self):
         self.assertRaises(
             ValueError,
             policy.DocumentedRuleDefault,
@@ -1523,24 +1560,25 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             operations=[{'path': '/v1/foos/', 'method': 'POST'}],
             deprecated_for_removal=True,
             deprecated_reason='Some reason.'
+            # no deprecated_since
         )
 
-    def test_deprecated_policy_must_include_deprecated_since(self):
+    def test_deprecated_policy_should_not_include_deprecated_meta(self):
         deprecated_rule = policy.DeprecatedRule(
             name='foo:bar',
             check_str='rule:baz'
         )
 
-        self.assertRaises(
-            ValueError,
-            policy.DocumentedRuleDefault,
-            name='foo:bar',
-            check_str='rule:baz',
-            description='Create a foo.',
-            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
-            deprecated_rule=deprecated_rule,
-            deprecated_reason='Some reason.'
-        )
+        with mock.patch('warnings.warn') as mock_warn:
+            policy.DocumentedRuleDefault(
+                name='foo:bar',
+                check_str='rule:baz',
+                description='Create a foo.',
+                operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+                deprecated_rule=deprecated_rule,
+                deprecated_reason='Some reason.'
+            )
+            mock_warn.assert_called_once()
 
     def test_deprecated_rule_requires_deprecated_rule_object(self):
         self.assertRaises(
@@ -1576,7 +1614,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
         # better.
         deprecated_rule = policy.DeprecatedRule(
             name='foo:bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N',
         )
         rule_list = [policy.DocumentedRuleDefault(
             name='foo:create_bar',
@@ -1584,8 +1624,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         self.enforcer.register_defaults(rule_list)
 
@@ -1610,7 +1648,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
         # better.
         deprecated_rule = policy.DeprecatedRule(
             name='foo:bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N',
         )
         rule_list = [policy.DocumentedRuleDefault(
             name='foo:create_bar',
@@ -1618,8 +1658,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         self.enforcer.register_defaults(rule_list)
 
@@ -1652,7 +1690,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
         # better.
         deprecated_rule = policy.DeprecatedRule(
             name='foo:bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N',
         )
         rule_list = [policy.DocumentedRuleDefault(
             name='foo:create_bar',
@@ -1660,8 +1700,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         self.enforcer.register_defaults(rule_list)
 
@@ -1696,7 +1734,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
         # Deprecate the policy name in favor of something better.
         deprecated_rule = policy.DeprecatedRule(
             name='old_rule',
-            check_str='role:bang'
+            check_str='role:bang',
+            deprecated_reason='"old_rule" is a bad name',
+            deprecated_since='N',
         )
         rule_list = [policy.DocumentedRuleDefault(
             name='new_rule',
@@ -1704,8 +1744,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Replacement for old_rule.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"old_rule" is a bad name',
-            deprecated_since='N'
         )]
         self.enforcer.register_defaults(rule_list)
 
@@ -1725,7 +1763,9 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
                                group='oslo_policy')
         deprecated_rule = policy.DeprecatedRule(
             name='foo:create_bar',
-            check_str='role:fizz'
+            check_str='role:fizz',
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N',
         )
 
         rule_list = [policy.DocumentedRuleDefault(
@@ -1734,8 +1774,6 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             description='Create a bar.',
             operations=[{'path': '/v1/bars', 'method': 'POST'}],
             deprecated_rule=deprecated_rule,
-            deprecated_reason='"role:bang" is a better default',
-            deprecated_since='N'
         )]
         enforcer = policy.Enforcer(self.conf)
         enforcer.register_defaults(rule_list)
@@ -1753,6 +1791,49 @@ class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
             enforcer.enforce('foo:create_bar', {}, {'roles': ['baz']})
         )
 
+    def test_deprecation_logic_is_only_performed_once_per_rule(self):
+        deprecated_rule = policy.DeprecatedRule(
+            name='foo:create_bar',
+            check_str='role:fizz'
+        )
+        rule = policy.DocumentedRuleDefault(
+            name='foo:create_bar',
+            check_str='role:bang',
+            description='Create a bar.',
+            operations=[{'path': '/v1/bars', 'method': 'POST'}],
+            deprecated_rule=deprecated_rule,
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N'
+        )
+        check = rule.check
+
+        enforcer = policy.Enforcer(self.conf)
+        enforcer.register_defaults([rule])
+
+        # Check that rule processing hasn't been done, yet
+        self.assertEqual({}, enforcer.rules)
+
+        # Load the rules
+        enforcer.load_rules()
+
+        # Loading the rules will store a version of the rule check string
+        # logically ORed with the check string of the deprecated value. Make
+        # sure this is happening but that the original rule check is unchanged
+        expected_check = policy.OrCheck([
+            _parser.parse_rule(cs) for cs in
+            [rule.check_str, deprecated_rule.check_str]
+        ])
+        self.assertIn('foo:create_bar', enforcer.rules)
+        self.assertEqual(
+            str(enforcer.rules['foo:create_bar']), str(expected_check))
+        self.assertEqual(check, rule.check)
+        # Hacky way to check whether _handle_deprecated_rule was called again.
+        # If a second call to load_rules doesn't overwrite our dummy rule then
+        # we know it didn't call the deprecated rule function again.
+        enforcer.rules['foo:create_bar'] = 'foo:bar'
+        enforcer.load_rules()
+        self.assertEqual('foo:bar', enforcer.rules['foo:create_bar'])
+
 
 class DocumentedRuleDefaultTestCase(base.PolicyBaseTestCase):
 
@@ -1837,6 +1918,18 @@ class DocumentedRuleDefaultTestCase(base.PolicyBaseTestCase):
                           operations=invalid_op)
 
 
+class DeprecatedRuleTestCase(base.PolicyBaseTestCase):
+
+    def test_should_include_deprecated_meta(self):
+        with mock.patch('warnings.warn') as mock_warn:
+            policy.DeprecatedRule(
+                name='foo:bar',
+                check_str='rule:baz'
+            )
+
+            mock_warn.assert_called_once()
+
+
 class EnforcerCheckRulesTest(base.PolicyBaseTestCase):
     def setUp(self):
         super(EnforcerCheckRulesTest, self).setUp()
@@ -1903,3 +1996,83 @@ class EnforcerCheckRulesTest(base.PolicyBaseTestCase):
         self.enforcer.load_rules(True)
 
         self.assertTrue(self.enforcer.check_rules())
+
+
+class PickPolicyFileTestCase(base.PolicyBaseTestCase):
+
+    def setUp(self):
+        super(PickPolicyFileTestCase, self).setUp()
+        self.data = {
+            'rule_admin': 'True',
+            'rule_admin2': 'is_admin:True'
+        }
+        self.tmpdir = self.useFixture(fixtures.TempDir())
+        original_search_dirs = cfg._search_dirs
+
+        def fake_search_dirs(dirs, name):
+            dirs.append(self.tmpdir.path)
+            return original_search_dirs(dirs, name)
+
+        mock_search_dir = self.useFixture(
+            fixtures.MockPatch('oslo_config.cfg._search_dirs')).mock
+        mock_search_dir.side_effect = fake_search_dirs
+
+        mock_cfg_location = self.useFixture(
+            fixtures.MockPatchObject(self.conf, 'get_location')).mock
+        mock_cfg_location.return_value = cfg.LocationInfo(
+            cfg.Locations.set_default, 'None')
+
+    def test_no_fallback_to_json_file(self):
+        tmpfilename = 'policy.yaml'
+        self.conf.set_override('policy_file', tmpfilename, group='oslo_policy')
+        jsonfile = os.path.join(self.tmpdir.path, 'policy.json')
+        with open(jsonfile, 'w') as fh:
+            jsonutils.dump(self.data, fh)
+
+        selected_policy_file = policy.pick_default_policy_file(
+            self.conf, fallback_to_json_file=False)
+        self.assertEqual(self.conf.oslo_policy.policy_file, tmpfilename)
+        self.assertEqual(selected_policy_file, tmpfilename)
+
+    def test_overridden_policy_file(self):
+        tmpfilename = 'nova-policy.yaml'
+        self.conf.set_override('policy_file', tmpfilename, group='oslo_policy')
+        selected_policy_file = policy.pick_default_policy_file(self.conf)
+        self.assertEqual(self.conf.oslo_policy.policy_file, tmpfilename)
+        self.assertEqual(selected_policy_file, tmpfilename)
+
+    def test_only_new_default_policy_file_exist(self):
+        self.conf.set_override('policy_file', 'policy.yaml',
+                               group='oslo_policy')
+        tmpfilename = os.path.join(self.tmpdir.path, 'policy.yaml')
+        with open(tmpfilename, 'w') as fh:
+            yaml.dump(self.data, fh)
+
+        selected_policy_file = policy.pick_default_policy_file(self.conf)
+        self.assertEqual(self.conf.oslo_policy.policy_file, 'policy.yaml')
+        self.assertEqual(selected_policy_file, 'policy.yaml')
+
+    def test_only_old_default_policy_file_exist(self):
+        self.conf.set_override('policy_file', 'policy.yaml',
+                               group='oslo_policy')
+        tmpfilename = os.path.join(self.tmpdir.path, 'policy.json')
+        with open(tmpfilename, 'w') as fh:
+            jsonutils.dump(self.data, fh)
+
+        selected_policy_file = policy.pick_default_policy_file(self.conf)
+        self.assertEqual(self.conf.oslo_policy.policy_file, 'policy.yaml')
+        self.assertEqual(selected_policy_file, 'policy.json')
+
+    def test_both_default_policy_file_exist(self):
+        self.conf.set_override('policy_file', 'policy.yaml',
+                               group='oslo_policy')
+        tmpfilename1 = os.path.join(self.tmpdir.path, 'policy.json')
+        with open(tmpfilename1, 'w') as fh:
+            jsonutils.dump(self.data, fh)
+        tmpfilename2 = os.path.join(self.tmpdir.path, 'policy.yaml')
+        with open(tmpfilename2, 'w') as fh:
+            yaml.dump(self.data, fh)
+
+        selected_policy_file = policy.pick_default_policy_file(self.conf)
+        self.assertEqual(self.conf.oslo_policy.policy_file, 'policy.yaml')
+        self.assertEqual(selected_policy_file, 'policy.yaml')
